// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include <iostream>
#include <cmath>
#include <dune/common/parallel/mpihelper.hh> // An initializer of MPI
#include <dune/common/exceptions.hh> // We use exceptions
#include <dune/grid/yaspgrid.hh>                // YetAnotherStructuredGrid
#include <dune/pdelab/finiteelementmap/qkfem.hh>
#include<dune/pdelab/backend/istl.hh>
#include <dune/pdelab/constraints/conforming.hh>
#include<dune/pdelab/function/callableadapter.hh>
#include <dune/geometry/referenceelements.hh>
#include<dune/grid/io/file/vtk.hh>
#include<dune/grid/io/file/vtk/vtkwriter.hh>
#include<dune/pdelab/common/vtkexport.hh>
#include<dune/pdelab/gridfunctionspace/vtk.hh>
#include  <dune/pdelab/gridfunctionspace/subspace.hh> //access subspace of PowerGFS

int main(int argc, char** argv) {
	Dune::MPIHelper &mpiHelper = Dune::MPIHelper::instance(argc, argv);

	const int dim = 2;

	using GridType = Dune::YaspGrid<dim>;
	using GridView = typename GridType::LeafGridView;
	using VTKWRITER =Dune::VTKWriter<GridView>;
	using VTKSEQUENCEWRITER =Dune::VTKSequenceWriter<GridView>;
	using FEM = Dune::PDELab::QkLocalFiniteElementMap<GridView, GridType::ctype,double,1>;
	using VBE0 = Dune::PDELab::ISTL::VectorBackend<>;

	//one component GridFunctionspace
	using GFS = Dune::PDELab::GridFunctionSpace<GridView, FEM, VBE0>;
	using ScalarGFS_RHS = Dune::PDELab::Backend::Vector<GFS,double>;

	//two component GridFunctionspace
	static const int GFSSIZE = 2;
	using PowerGFS = Dune::PDELab::PowerGridFunctionSpace<GFS,GFSSIZE,VBE0,Dune::PDELab::EntityBlockedOrderingTag>;
	using PowerGFS_RHS = Dune::PDELab::Backend::Vector<PowerGFS,double>;

	//build the grid
	int overlap = 1;
	std::bitset<dim> B(false);
	Dune::FieldVector<double, dim> domain(1.0);
	std::array<int, dim> domainDims;
	domainDims[0] = 20;
	domainDims[1] = 20;
	std::shared_ptr<GridType> grid_p = std::shared_ptr<GridType>(new GridType(domain, domainDims, B, overlap, Dune::MPIHelper::getCollectiveCommunication()));
	std::shared_ptr<GridView> gv_p = std::shared_ptr<GridView>(std::make_shared<GridView>(grid_p->leafGridView()));

	//build FEM and gridfunctionspace
	std::shared_ptr<FEM> fem_p = std::shared_ptr<FEM>(new FEM(*gv_p));
	std::shared_ptr<GFS> gfs_p = std::shared_ptr<GFS>(new GFS(*gv_p, *fem_p));
	std::shared_ptr<PowerGFS> powerGFS_p = std::shared_ptr<PowerGFS>(new PowerGFS(*gfs_p));

	//build two component rhs vector z_p
	std::shared_ptr<PowerGFS_RHS> z_p = std::shared_ptr<PowerGFS_RHS>(new PowerGFS_RHS(*powerGFS_p));

	//interpolate dummy function into array: z_p
	const double pi = std::acos(-1);
	auto glambda = [&](const auto& global_coord) {
		if constexpr (GFSSIZE == 2) {
			Dune::FieldVector<double,2> rv(0.0);
			rv[0] = std::sin(global_coord[0]*(0.5*pi))*std::sin(global_coord[1]*(0.5*pi));
			rv[1] = rv[0];
			return rv;
		} else {
			return std::sin(global_coord[0]*(0.5*pi))*std::sin(global_coord[1]*(0.5*pi));
		}
	};
	auto g = Dune::PDELab::makeGridFunctionFromCallable(*gv_p, glambda);
	Dune::PDELab::interpolate(g, *powerGFS_p, *z_p);

	//prepare output writer DGF Types:

	//for field:
	using VECTOR_DGF_1 = Dune::PDELab::VectorDiscreteGridFunction<PowerGFS, PowerGFS_RHS,1>;
	//force size 1 -->FieldVector<double,1> not double!
	//using VECTOR_DGF_2 = Dune::PDELab::VectorDiscreteGridFunction<PowerGFS, PowerGFS_RHS,2>;	//force size 2

	using VECTOR_VTK_GF_ADAPTER = Dune::PDELab::VTKGridFunctionAdapter<VECTOR_DGF_1>;

	using ScalarDGF = Dune::PDELab::DiscreteGridFunction<GFS, ScalarGFS_RHS>;
	using SCALAR_VTK_GF_ADAPTER = Dune::PDELab::VTKGridFunctionAdapter<ScalarDGF>;

	//GridFunciton gradient types:
	using DGF_GRADIENT = Dune::PDELab::DiscreteGridFunctionGradient<GFS,PowerGFS_RHS>;
	using VECTOR_DGF_GRADIENT = Dune::PDELab::VectorDiscreteGridFunctionGradient<PowerGFS,PowerGFS_RHS>;
	using VECTOR_DGF_GRADIENT_ADAPTER = Dune::PDELab::VTKGridFunctionAdapter<VECTOR_DGF_GRADIENT>;

	//GridFunction Adapters:
	using VTKFG = Dune::PDELab::VTKGridFunctionAdapter<DGF_GRADIENT>;

	//create writer:
	VTKWRITER vtkwriter(*gv_p);
	std::string filename("output");
	VTKSEQUENCEWRITER vtkSequenceWriter(std::make_shared<VTKWRITER>(vtkwriter), filename, "./", "");

	//build two one-component dgfs!
	std::shared_ptr<VECTOR_DGF_1> dgf_z1_p = std::shared_ptr<VECTOR_DGF_1>(new VECTOR_DGF_1(*powerGFS_p, *z_p, Dune::FieldVector<int, 1>(0)));	//take first component!
	std::shared_ptr<VECTOR_DGF_1> dgf_z2_p = std::shared_ptr<VECTOR_DGF_1>(new VECTOR_DGF_1(*powerGFS_p, *z_p, Dune::FieldVector<int, 1>(1)));	//take second component!
	//add them to the writer:
	vtkSequenceWriter.addVertexData(std::shared_ptr<VECTOR_VTK_GF_ADAPTER>(new VECTOR_VTK_GF_ADAPTER(*dgf_z1_p, "z_1")));
	vtkSequenceWriter.addVertexData(std::shared_ptr<VECTOR_VTK_GF_ADAPTER>(new VECTOR_VTK_GF_ADAPTER(*dgf_z2_p, "z_2")));

	//I want to write the gradient of z_1!

	//Option 1: build a vector DGF filled with gradients:------------------------------------------------------
	//Problem: return type of the DGF is not scalar, so the adapter fails
//	std::shared_ptr<VECTOR_DGF_GRADIENT> vec_DGF_1_grad_p = std::shared_ptr<VECTOR_DGF_GRADIENT>(new VECTOR_DGF_GRADIENT(*gfs_p, *z_p));
//	vtkSequenceWriter.addVertexData(std::shared_ptr<VECTOR_DGF_GRADIENT_ADAPTER>(new VECTOR_DGF_GRADIENT_ADAPTER(*vec_DGF_1_grad_p, "z1_grad")));
	//Error: error: cannot convert ‘Dune::DenseMatVecTraits<Dune::FieldMatrix<double, 2, 2> >::row_type {aka Dune::FieldVector<double, 2>}’ to ‘double’ in return

	//Option 2: extract subspace, to try to use the one-component dgf adapters..
//	using Z1_SUBSPACE = typename Dune::PDELab::GridFunctionSubSpace<PowerGFS,Dune::TypeTree::TreePath<0>>;
//	auto z2_gfs_p = std::make_shared<Z1_SUBSPACE>(*powerGFS_p);
//	using Z1_SUBSPACE_RHS =Dune::PDELab::Backend::Vector<Z1_SUBSPACE,double>;
//	using Z1_SUBSPACE_DGF_GRADIENT = Dune::PDELab::DiscreteGridFunctionGradient<Z1_SUBSPACE,PowerGFS_RHS>;
//	std::shared_ptr<Z1_SUBSPACE_DGF_GRADIENT> subspace_dgf_gradient_p = std::shared_ptr<Z1_SUBSPACE_DGF_GRADIENT>(new Z1_SUBSPACE_DGF_GRADIENT(*z2_gfs_p, *z_p));
//	vtkSequenceWriter.addVertexData(std::shared_ptr<SCALAR_VTK_GF_ADAPTER>(new SCALAR_VTK_GF_ADAPTER(*subspace_dgf_gradient_p, "z1_grad")));

	//Option 3: create adapter from subspace type:
//	using Z1_SUBSPACE = typename Dune::PDELab::GridFunctionSubSpace<PowerGFS,Dune::TypeTree::TreePath<0>>;
//	auto z2_gfs_p = std::make_shared<Z1_SUBSPACE>(*powerGFS_p);
//	using Z1_SUBSPACE_RHS =Dune::PDELab::Backend::Vector<Z1_SUBSPACE,double>;
//	using Z1_SUBSPACE_DGF_GRADIENT = Dune::PDELab::DiscreteGridFunctionGradient<Z1_SUBSPACE,PowerGFS_RHS>;
//	using SUBSPACE_VTK_GF_ADAPTER = Dune::PDELab::VTKGridFunctionAdapter<Z1_SUBSPACE_DGF_GRADIENT>;
//	std::shared_ptr<SUBSPACE_VTK_GF_ADAPTER> subspace_dgf_gradient_p = std::shared_ptr<SUBSPACE_VTK_GF_ADAPTER>(new SUBSPACE_VTK_GF_ADAPTER(*z2_gfs_p, *z_p));
//	vtkSequenceWriter.addVertexData(std::shared_ptr<SCALAR_VTK_GF_ADAPTER>(new SCALAR_VTK_GF_ADAPTER(*subspace_dgf_gradient_p, "z1_grad")));


	vtkSequenceWriter.write(0.0, Dune::VTK::appendedraw);
	return 0;
}
